/*********************************************************************/
/* File: screen.c - functions for the screen handling                */
/*                            HUT TINTIN                             */
/*                     coded by Ossi Tulonen                         */
/*********************************************************************/
/*
 * Copyright (c) 1992 1993 1994 Ossi tulonen
 * All rights reserved.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 */
 
#include <stdio.h>
#ifdef TERMIOS
#  ifdef SUN
extern "C" {
#  endif
#  include <termios.h>
#  ifdef SUN
}
#  endif
#  include <sys/ioctl.h>
#else
#  include <sys/ioctl.h>
#  ifdef DIRTY_REDEFINE
#    undef TIOCSETP
#    define TIOCSETP 0x80067409
#  endif
#  include <sgtty.h>
#endif
#ifndef NSTDLIBS
#  include <stdlib.h>
#endif
#ifdef OSF1v3
typedef int clockid_t;
#endif
#ifdef TERMCAP
#  if defined(__linux__)
#    include <termcap.h>
#  else
extern "C" {
  int tgetent(void *, const char *);
  int tgetnum(const char *);
  char *tgetstr(const char *, char **);
  char *tgoto(const char *, int, int);
}
#  endif
#endif

#include <time.h>
#include <ctype.h>
#include <string.h>
#include <assert.h>

#include "structs.h"
#include "color.h"
#include "history.h"
#include "llist.h"
#include "main.h"
#include "parse.h"
#include "screen.h"
#include "utility.h"

#define INPUTBUF master_table.inputbuf
#define INPUTPOS master_table.inputbuf_pos
#define CURSOR   master_table.cursor_pos

unsigned char window::Scrcolormode[4] = { 0x01, 0, 0, 0};
int window::ScrRows = 0;
int window::ScrCols = 0;
scrollsize window::ScrScroll = ((scrollsize){1, 1});
position window::Scrcursor   = ((position){0, 0});
window *window::windows      = 0;
window *window::active       = 0;

const uchar *CEL = 0;
const uchar *CSL = 0;
const uchar *CL  = 0;
const uchar *CES = 0;
const uchar *CSS = 0;
const uchar *CLEARSCREEN = 0;
const uchar *AUTOWARPON  = 0;
const uchar *AUTOWARPOFF = 0;
const uchar *HOME     = 0;
const uchar *CUP      = 0;
const uchar *CDOWN    = 0;
const uchar *FORWARD  = 0;
const uchar *BACKWARD = 0;
const uchar *RESETCOLOR = 0;
const uchar *MAXSCROLLSIZE = 0;
const uchar *INSERTLINE = 0;
const uchar *VBELL    = 0;
const uchar *MOVE     = 0;

int TABSIZE     = 0;

extern int default_chars_per_line;
extern int default_lines_in_scr;
extern struct master_table_data master_table;
#ifdef TERMIOS
extern struct termios original_state;
#else
extern struct sgttyb original_state;
extern struct tchars original_chars;
#endif

#define CUR_STOP 5

// **********************
// class window functions
// **********************

window::window()  // this is first window
{
  next = 0;
  window::windows = this;
  
  name = (uchar *)strdup("main");
  colormode[0] = 0x01;
  colormode[1] = 0;
  colormode[2] = 0;
  colormode[3] = 0;
  
  WinSize.start = 1;
  WinSize.end   = window::ScrRows;
  Scroll.start = 1;
  Scroll.end   = window::ScrRows;
  cursor.col   = 1;
  cursor.row   = 1;
  timedelay    = -1;
  ptexrem      = FALSE;
  autowarp     = TRUE;
  window_hist_pos = 0;
  window_hist_on  = 0;
  window_hist_clip = FALSE;
  
  active_ses   = 0;

  lastupdate = 0;

  lastactive     = this;
  window::active = this;
  clear();
}

window::window(const unsigned char *nam, const int size, window *win)
{
  window *pre;
  session *mses, *ses, *nextses;

  assert(size <= win->WinSize.end - win->WinSize.start - 2);
  
  name = (uchar *)strdup((const char*)nam);
  
  colormode[0] = 0x01;
  colormode[1] = 0;
  colormode[2] = 0;
  colormode[3] = 0;

  WinSize.start      = win->WinSize.start;
  win->WinSize.start = WinSize.start + size;

  WinSize.end        = WinSize.start + size - 1;

  win->Scroll.start  = win->WinSize.start;
  
  Scroll.start = WinSize.start;
  Scroll.end   = WinSize.end;
  cursor.col   = 1;
  cursor.row   = WinSize.start;
  timedelay    = -1;
  ptexrem      = FALSE;
  autowarp     = TRUE;
  window_hist_pos = 0;
  window_hist_on  = 0;
  window_hist_clip = FALSE;

  mses = master_table.main_session;
  active_ses   = mses;
  if(mses->get_input()->get_active_ses() == mses) {
    for(ses = session::get_sessions(); ses; ses = ses->get_next())
      if(ses->get_input() == mses->get_input())
	break;
    
    mses->get_input()->set_active_ses(ses);
  }
  mses->set_input(this);
  mses->set_output(this);
  
  lastupdate = 0;
  
  if(window::windows == win) {
    next = window::windows;
    window::windows = this;
  } else {
    for(pre = window::windows; pre && pre->next != win; pre = pre->next)
      ;
    assert(pre);
    next = pre->next;
    pre->next = this;
  }
  lastactive     = this;
  window::active = this;
  redraw();
  win->redraw();
  for(mses = session::get_sessions(); mses; mses = nextses) {
    nextses = mses->get_next();
    if(mses->get_output() == this || mses->get_output() == win) 
      mses->send_naws();
  }
}

window::~window()
{
  int tmp;
  window *pre = 0, *update;
  session *ses, *newactiv, *nextses;
  
  if(this == window::windows) {
    window::windows = next;
  } else {
    for(pre = windows; pre && pre->next != this; pre = pre->next)
      ;
    pre->next = next;
  }
  if(name)
    free(name);
  
  if(master_table.output_window == this)
    master_table.output_window = 0;
  
  if(!window::windows)
    exit_tintin(0);
  
  if(window::active == this) {
    window::active = window::windows;
    window::active->set_ptexrem();
  }

  update = 0;
  if(next) {
    next->WinSize.start = WinSize.start;
    next->Scroll.start  = WinSize.start;
    update = next;
  } else {
    tmp = pre->WinSize.end - pre->Scroll.end;
    pre->WinSize.end = WinSize.end;
    pre->Scroll.end  = WinSize.end - tmp;
    update = pre;
  }
  
  if(master_table.main_window == this)
    master_table.main_window = window::active;
  
  for(ses = session::get_sessions(); ses; ses = ses->get_next()) {
    if(ses->get_input() == this)
      ses->set_input(window::active);

    if(ses->get_output() == this)
      ses->set_output(window::active);
  }
  if(master_table.main_session) {
    if(master_table.main_session->get_input() == this)
      master_table.main_session->set_input(window::active);
    
    if(master_table.main_session->get_output() == this)
      master_table.main_session->set_output(window::active);
  }
  
  if(!window::active->active_ses) {
    for(newactiv = session::get_sessions();
	newactiv;
	newactiv = newactiv->get_next())
      if(newactiv->get_input() == window::active)
	break;
    if(newactiv)
      window::active->active_ses = newactiv;
    else {
      if((newactiv = master_table.main_session)) {
	if(newactiv->get_input()->get_active_ses() == newactiv) {
	  for(ses = session::get_sessions(); ses; ses = ses->get_next())
	    if(ses->get_input() == newactiv->get_input())
	      break;
	  
	  newactiv->get_input()->set_active_ses(ses);
	}
	window::active->active_ses = newactiv;
	newactiv->set_input(window::active);
      }
    }
  }
  for(ses = session::get_sessions(); ses; ses = nextses) {
    nextses = ses->get_next();
    if(ses->get_input() == this || ses->get_output() == update)
      ses->send_naws();
  }
}

void window::move(const position *pos)
{
  assert(pos->row >= WinSize.start && pos->row <= WinSize.end);
  assert(pos->col >= 0 && pos->col <= window::ScrCols);
  
#ifdef TERMCAP
  if(MOVE) {
    printf("%s", tgoto((const char *)MOVE, pos->col - 1, pos->row - 1));
  } else {
#endif
    printf("\033[%d;%dH", pos->row, pos->col);
#ifdef TERMCAP
  }
#endif
  Scrcursor = *pos;
  cursor = *pos;
}

void window::move(const int row, const int col)
{
  assert(row >= WinSize.start && row <= WinSize.end);
  assert(col >= 0 && col <= window::ScrCols);

#ifdef TERMCAP
  if(MOVE) {
    printf("%s", tgoto((const char *)MOVE, col - 1, row - 1));
  } else {
#endif
    printf("\033[%d;%dH", row, col);
#ifdef TERMCAP
  }
#endif
  Scrcursor.row = row;
  Scrcursor.col = col;
  cursor.row = row;
  cursor.col = col;
}

void window::moveforward()
{
  check_color();

  if(++cursor.col > window::ScrCols) {
    if(autowarp) {
      cursor.col = 1;
      if(cursor.row < Scroll.end)
	cursor.row++;
      move(&cursor);
    }
  } else {
    if(window::Scrcursor.col != cursor.col - 1 ||
       window::Scrcursor.row != cursor.row) {
      move(&cursor);
    } else {
      printf("%s", FORWARD);
      window::Scrcursor.col = cursor.col;
    }
  }
}

void window::movebackward()
{
  check_color();
  if(!(--cursor.col)) {
    cursor.col = window::ScrCols;
    if(cursor.row > Scroll.start)
      cursor.row--;
    move(&cursor);
  } else {
    if(window::Scrcursor.col != cursor.col + 1 ||
       window::Scrcursor.row != cursor.row) {
      move(&cursor);
    } else {
      printf("%s", BACKWARD);
      window::Scrcursor.col = cursor.col;
    }
  }
}

void window::scroll_region(const scrollsize *screg)
{
  assert(screg->start >= WinSize.start &&
	 screg->end <= WinSize.end - get_time_on());
  
  printf("\033[%d;%dr", screg->start, screg->end);
  window::ScrScroll = *screg;
  move(&cursor);
}

void window::scroll_region(const int start, const int end)
{
  assert(start >= WinSize.start && end <= WinSize.end - get_time_on());
  
  printf("\033[%d;%dr", start, end);
  window::ScrScroll.start = start;
  window::ScrScroll.end   = end;
  move(&cursor);
}

void window::clear()
{
  int i, WinEnd;
  position pos;
  
  WinEnd = WinSize.end;
  if(get_time_on())
    WinEnd--;
  
  check_color();
  if(WinSize.start == 1 && WinEnd == window::ScrRows) {
    printf("%s", CLEARSCREEN);
    cursor.row = 1;
    cursor.col = 1;
    Scrcursor.row = 1;
    Scrcursor.col = 1;
  } else if(WinEnd == window::ScrRows) {
    move(WinSize.start, 1);
    printf("%s", CES);
  } else {
    pos.row = WinSize.start;
    pos.col = 1;
    move(&pos);
    
    printf("%s", CEL);
    
    if(WinSize.start != WinEnd) {
      for(i = WinSize.start + 1; i <= WinEnd; i++) {
	move(i, 1);
	printf("%s", CEL);
      }
      move(&pos);
    }
  }
  ptexrem = TRUE;
  if(window::active == this)
    master_table.utexrem = TRUE;
  
  fflush(stdout);
}

void window::clearend()
{
  int i, WinEnd;
  position pos;
  
  WinEnd = WinSize.end;
  if(get_time_on())
    WinEnd--;
  
  if(!window_hist_on) {
    if(window::Scrcursor.col != cursor.col ||
       window::Scrcursor.row != cursor.row)
      move(&cursor);
    
    check_color();
     
    if(WinEnd == window::ScrRows) {
      printf("%s", CES);
    } else {
      printf("%s", CEL);
      if(cursor.row != WinEnd) {
	pos = cursor;
	for(i = cursor.row + 1; i <= WinEnd; i++) {
	  move(i, 1);
	  printf("%s", CEL);
	}
	move(&pos);
      }
    }
  }
  fflush(stdout);
}

void window::clearstart()
{
  int i;
  position pos;
  
  if(!window_hist_on) {
    if(window::Scrcursor.col != cursor.col ||
       window::Scrcursor.row != cursor.row)
      move(&cursor);
    
    check_color();
 
    if(WinSize.start == 1) {
      printf("%s", CSS);
    } else {
      printf("%s", CSL);
      if(cursor.row > WinSize.start) {
	pos = cursor;
	for(i = cursor.row - 1; i >= WinSize.start; i--) {
	  move(i, 1);
	  printf("%s", CEL);
	}
	move(&pos);
      }
    }
  }
  fflush(stdout);
}

void window::redraw(const int position)
{
  int i = 0, num, pos, histpos, maxlines, startline, lines, tmp, line;
  session *ses = 0;
  window *win;
  struct listnode *ln;
  struct position orgpos;
  
  orgpos = cursor;
  clear();
  if(ScrScroll.start != Scroll.start || ScrScroll.end != Scroll.end) {
    scroll_region(&Scroll);
  }
 
  if(window_hist_on) {
    ses = window_hist_on;
    
    num = ses->get_hist_num();
    
    maxlines = WinSize.end - WinSize.start;
    if(get_time_on())
      maxlines--;
    startline = WinSize.start;
    
    if(num > maxlines)
      num = maxlines;
    
    lines = maxlines;
    if(lines > ses->get_hist_num())
      lines = ses->get_hist_num();

    tmp = (WinSize.start + (lines - 1) - orgpos.row);
    if(tmp < 0) {
      orgpos.col = 1;
      tmp = 0;
    }
    pos = window_hist_pos + tmp + 1;   // cursor position
    
    if(position < 0) {
      histpos = pos;
    } else if(position == 0) {
      histpos = pos + maxlines / 2;
    } else {
      histpos = pos + maxlines - 1;
    }

    if(histpos < num)
      histpos = num;
    
    if(histpos > ses->get_hist_num())
      histpos = ses->get_hist_num();
    
    orgpos.row = WinSize.start + histpos - pos;

    window_hist_pos = histpos - num;
    
    for(i = 0; i < num; i++) {
      move(startline + i, 1);    /* cursor to right position */
      
      line = ses->get_hist_end() - histpos--;
      if(line < 0)
	line += MAX_SES_HIST;
      
      printf("%s", ses->get_scrhistory(line));
    }
    
    if(ses->get_hist_markp() > window_hist_pos &&
       ses->get_hist_markp() <= window_hist_pos + num) {
      
      move(WinSize.start + window_hist_pos + num - ses->get_hist_markp(),
	   ses->get_hist_markc());
      
      if((ln = searchnode_list(ses->colors, (unsigned char *)"HISTMARK"))) {
	printf("%s", ln->right);
	color_to_bits(ln->right, window::Scrcolormode);
      } else {
	printf("%s", TIMELINEON);
	color_to_bits(TIMELINEON, window::Scrcolormode);
      }
      
      line = ses->get_hist_end() - ses->get_hist_markp();
      if(line < 0)
	line += MAX_SES_HIST;
      
      printf("%c", *(ses->get_scrhistory(line) + ses->get_hist_markc() - 1));
    }
    
    cursor.row = orgpos.row;
    cursor.col = orgpos.col;
    print_win_hist_line(TRUE);
    ptexrem = FALSE;
  } else {
    for(win = window::windows; win; win = win->next) {
      if(win->get_active_ses() &&
	 win->get_active_ses()->get_output() == this) {
	ses = win->get_active_ses();
	break;
      }
    }
  
    if(!ses) {
      for(ses = master_table.main_session; ses; ses = ses->get_next())
	if(ses->get_output() == this)
	  break;
    }

    if(ses) {
      num = ses->get_hist_num();

      maxlines = Scroll.end - Scroll.start;
      startline = Scroll.start;
      
      if(num > maxlines)
	num = maxlines;
      
      histpos = window_hist_pos + num;
      
      if(histpos > ses->get_hist_num())
	histpos = ses->get_hist_num();
      
      for(i = 0; i < num; i++) {
	move(startline + i, 1);    /* cursor to right position */
	
	pos = ses->get_hist_end() - histpos--;
	if(pos < 0)
	  pos += MAX_SES_HIST;
	
	printf("%s", ses->get_scrhistory(pos));
      }
    }
    move(Scroll.start + i, 1);    /* cursor to right position */
    print_prompt(TRUE, FALSE);
  }
  print_time(TRUE);
}

void window::bell()
{
  print_time(FALSE, TRUE);
  putchar('\a');
  if(VBELL)
    printf("%s", VBELL);
  
  fflush(stdout);
}

void window::print(const unsigned char *cptr)
{
  unsigned char buf[BUFFER_SIZE + 100], *str, tmpbuf[100], *tmp;
  unsigned char numbuf[100], *num, numbuf2[100], *num2;
  int showlen = 0, i, j, decmode, WinEnd;
  position orgcursor;

  if(window::ScrScroll.start != Scroll.start ||
     window::ScrScroll.end   != Scroll.end) {
    scroll_region(&Scroll);
  }
  
  if(window::Scrcursor.col != cursor.col ||
     window::Scrcursor.row != cursor.row) {
    move(&cursor);
  }
  
  check_color();
   
  WinEnd = WinSize.end;
  if(get_time_on())
    WinEnd--;
  
  str = buf;
  while(*cptr) {
    showlen = window::Scrcursor.col - 1;
    while((showlen < window::ScrCols || !autowarp) && *cptr) {
      if(*cptr == ESC) {
	*str = '\0';
	cursor.col = showlen + 1;
	if(cursor.col > window::ScrCols)
	  cursor.col = window::ScrCols;

	printf("%s", buf);
	Scrcursor.col = cursor.col;
	str = buf;
	
	tmp = tmpbuf;
	*tmp++ = *cptr++;
	if(*cptr == '[') {
	  *tmp++ = *cptr++;
	  
	  decmode = FALSE;
	  if(*cptr == '?') {
	    decmode = TRUE;
	    *tmp++ = *cptr++;
	  }
	  
	  num = numbuf;
	  while(isdigit(*cptr)) {
	    *num++ = *cptr;
	    *tmp++ = *cptr++;
	  }
	  *num = '\0';
	  
	  num2 = numbuf2;
	  if(*cptr == ';') {
	    *tmp++ = *cptr++;
	    while(isdigit(*cptr)) {
	      *num2++ = *cptr;
	      *tmp++  = *cptr++;
	    }
	  }
	  *num2 = '\0';
	  while(*cptr == ';') {
	    *tmp++ = *cptr++;
	    while(isdigit(*cptr)) {
	      *num2  = *cptr;
	      *tmp++ = *cptr++;
	    }
	  }
	  *tmp++ = *cptr;
	  *tmp = '\0';
	  
	  if(decmode && *cptr != 'J' && *cptr != 'K') {
	    switch(*cptr) {
	    case 'l':
	      if(*numbuf)
		i = atoi((const char *)numbuf);
	      else
		i = 0;
	      switch(i) {
	      case 0:                   // IGNORED
		break;
	      case 7:                   // AUTOWARP OFF
		autowarp = FALSE;
		break;
	      }
#ifdef DEBUG
	      if(*numbuf2) {
		if(master_table.main_session->logfile)
		  fprintf(master_table.main_session->logfile,
			  "UNSUPORTED multi vt100-code ESC[?;;l\n");
		else
		  fprintf(stderr, "UNSUPORTED multi vt100-code ESC[?;;l\n");
	      }
#endif
	      cptr++;
	      break;
	    case 'h':
	      if(*numbuf)
		i = atoi((const char *)numbuf);
	      else
		i = 0;
	      switch(i) {
	      case 0:                   // IGNORED
		break;
	      case 7:                   // AUTOWARP ON
		autowarp = TRUE;
		break;
	      }
#ifdef DEBUG
	      if(*numbuf2) {
		if(master_table.main_session->logfile)
		  fprintf(master_table.main_session->logfile,
			  "UNSUPORTED multi vt100-code ESC[;;h\n");
		else
		  fprintf(stderr, "UNSUPORTED multi vt100-code ESC[;;h\n");
	      }
#endif
	      cptr++;
	      break;
	    default:
#ifdef DEBUG
	      *tmpbuf = 'C';
	      if(master_table.main_session->logfile)
		fprintf(master_table.main_session->logfile,
			"UNSUPORTED ansi-code ES%s\n", tmpbuf);
	      else
		fprintf(stderr, "UNSUPORTED ansi-code ES%s\n", tmpbuf);
#endif
	      cptr++;
	      break;
	    }  /* switch */
	  } else {
	    switch(*cptr) {
	    case 'm':
	      printf("%s", tmpbuf);
	      color_to_bits(tmpbuf, colormode);
	      window::Scrcolormode[0] = colormode[0];
	      window::Scrcolormode[1] = colormode[1];
	      window::Scrcolormode[2] = colormode[2];
	      window::Scrcolormode[3] = colormode[3];
	      cptr++;
	      break;
	    case 'r':
	      if(*numbuf) {
		i = atoi((const char *)numbuf);
		if(i < 1)
		  i = 1;
		else if(i > WinEnd - WinSize.start + 1)
		  i = WinEnd - WinSize.start + 1;
	      } else
		i = 1;
	      
	      if(*numbuf2) {
		j = atoi((const char *)numbuf2);
		if(j < 1)
		  j = 1;
		else if(j > WinEnd - WinSize.start + 1)
		  j = WinEnd - WinSize.start + 1;
	      } else
		j = WinEnd - WinSize.start + 1;
	      
	      Scroll.start = WinSize.start + i - 1;
	      Scroll.end   = WinSize.start + j - 1;
	      
	      scroll_region(&Scroll);
	      cptr++;
	      break;
	    case 'A':
	      if(*numbuf)
		i = atoi((const char *)numbuf);
	      else
		i = 1;
	      
	      if(cursor.row - i < WinSize.start)
		i = cursor.row - WinSize.start;
	      
	      if(i > 0) {
		cursor.row -= i;
		Scrcursor.row = cursor.row;
		if(i == 1)
		  printf("%s", CUP);
		else
		  printf("\033[%dA", i);
	      }
	      cptr++;
	      break;
	    case 'B':
	      if(*numbuf)
		i = atoi((const char *)numbuf);
	      else
		i = 1;
	      
	      if(cursor.row + i > WinEnd)
		i = WinEnd - cursor.row;
	      
	      if(i > 0) {
		cursor.row += i;
		Scrcursor.row = cursor.row;
		if(i == 1)
		  printf("%s", CDOWN);
		else
		  printf("\033[%dB", i);
	      }
	      cptr++;
	      break;
	    case 'C':
	      if(*numbuf)
		i = atoi((const char *)numbuf);
	      else
		i = 1;
	      if(showlen + i > window::ScrCols)
		i = window::ScrCols - showlen;
	      
	      if(i > 0) {
		showlen += i;
		if(i == 1)
		  printf("%s", FORWARD);
		else
		  printf("\033[%dC", i);
	      }
	      cptr++;
	      break;
	    case 'D':
	      if(*numbuf)
		i = atoi((const char *)numbuf);
	      else
		i = 1;
	      if(showlen - i < 0)
		i = showlen;
	      
	      if(i > 0) {
		showlen -= i;
		if(i == 1)
		  printf("%s", BACKWARD);
		else
		  printf("\033[%dD", i);
	      }
	      cptr++;
	      break;
	    case 'H':
	      if(*numbuf) {
		i = atoi((const char *)numbuf);
		if(i <= 0)
		  i = 1;
		else if(i > WinEnd - WinSize.start + 1)
		  i = WinEnd - WinSize.start + 1;
	      } else
		i = 1;
	      
	      if(*numbuf2) {
		j = atoi((const char *)numbuf2);
		if(j <= 0)
		  j = 1;
		else if(j > window::ScrCols)
		  j = window::ScrCols;
	      } else
		j = 1;
	      
	      move(WinSize.start + i - 1, j);
	      showlen = window::Scrcursor.col - 1;
	      cptr++;
	      break;
	    case 'J':
	      if(!*numbuf || (*numbuf == '0' && !*(numbuf + 1))) {
		clearend();
	      } else if(*numbuf == '1' && !*(numbuf + 1)) {
		clearstart();
	      } else if(*numbuf == '2' && !*(numbuf + 1)) {
		orgcursor = cursor;
		clear();
		
		move(&orgcursor);
		showlen = window::Scrcursor.col - 1;
	      } else {
#ifdef DEBUG
		*tmpbuf = 'C';
		if(master_table.main_session->logfile)
		  fprintf(master_table.main_session->logfile,
			  "UNSUPORTED ansi-code ES%s\n", tmpbuf);
		else
		  fprintf(stderr, "UNSUPORTED ansi-code ES%s\n", tmpbuf);
#endif
	      }
	      cptr++;
	      break;
	    case 'K':
	      if(!*numbuf || (i = atoi((const char *)numbuf) == 0)) {
		printf("%s", CEL);
	      } else if(i = 1) {
		printf("%s", CSL);
	      } else if(i == 2) {
		printf("%s", CL);
	      } else {
#ifdef DEBUG
		*tmpbuf = 'C';
		if(master_table.main_session->logfile)
		  fprintf(master_table.main_session->logfile,
			  "UNSUPORTED ansi-code ES%s\n", tmpbuf);
		else
		  fprintf(stderr, "UNSUPORTED ansi-code ES%s\n", tmpbuf);
#endif
	      }
	      cptr++;
	      break;
	    case 'L':
	      if(*numbuf)
		printf("\033[%sL", numbuf);
	      else
		printf("%s", INSERTLINE);
	      cptr++;
	      break;
	    default:
#ifdef DEBUG
	      *tmpbuf = 'C';
	      if(master_table.main_session->logfile)
		fprintf(master_table.main_session->logfile,
			"UNSUPORTED ansi-code ES%s\n", tmpbuf);
	      else
		fprintf(stderr, "UNSUPORTED ansi-code ES%s\n", tmpbuf);
#endif
	      cptr++;
	      break;
	    }  /* switch */
	  }
 	} else {
	  switch(*cptr) {
	  case '=':                // numeric keypad application mode
	    printf("\033=");
	    break;
	  case '>':                // numeric keypad numeric mode
	    printf("\033>");
	    break;
	  case 'D':
	    printf("\033D");
	    if(cursor.row < Scroll.end) {
	      cursor.row++;
	      Scrcursor.row++;
	    }
	    break;
	  case 'M':
	    printf("\033M");
	    if(cursor.row > Scroll.start) {
	      cursor.row--;
	      Scrcursor.row--;
	    }
	    break;
	  default:
#ifdef DEBUG
	    if(master_table.main_session->logfile)
	      fprintf(master_table.main_session->logfile,
		      "UNSUPORTED ESC code ESC%c\n", *cptr);
	    else
	      fprintf(stderr, "UNSUPORTED ESC code ESC%c\n", *cptr);
#endif
	    break;
	  } /* switch */
	  cptr++;
	}
      } else {
	switch(*cptr) {
	case '\b':                       // backspace
	  if(showlen > 0) {
	    showlen--;
	    *str++ = *cptr;
	  }
	  cptr++;
	  break;
	case '\t':                       // tab
	  *str = '\0';
	  cursor.col = showlen + 1;
	  if(cursor.col > window::ScrCols)
	    cursor.col = window::ScrCols;
 
	  printf("%s", buf);
	  Scrcursor.col = cursor.col;
	  str = buf;
	  
	  showlen++;
	  if(showlen % TABSIZE)
	    showlen += TABSIZE - (showlen % TABSIZE);
	  
	  if(showlen + 1 > window::ScrCols)
	    showlen = window::ScrCols;
	  
	  move(cursor.row, showlen);
	  cptr++;
	  break;
	case '\a':                       // bell
	  bell();
	  cptr++;
	  break;
	case '\r':
	  showlen = 0;
	  *str++ = '\r';
	  cptr++;
	  break;
	case '\f':
	  *str = '\0';
	  cursor.col = showlen + 1;
	  if(cursor.col > window::ScrCols)
	    cursor.col = window::ScrCols;
 
	  printf("%s\f", buf);
	  Scrcursor.col = cursor.col;
	  str = buf;
	  
	  if(cursor.row < Scroll.end) {
	    cursor.row++;
	    Scrcursor.row++;
	  }
	  cptr++;
	  break;
	case '\n':
	  *str = '\0';
	  cursor.col = showlen + 1;
	  if(cursor.col > window::ScrCols)
	    cursor.col = window::ScrCols;
	  
	  printf("%s\n", buf);
	  Scrcursor.col = cursor.col;
	  str = buf;
	  
	  if(cursor.row < Scroll.end) {
	    cursor.row++;
	    Scrcursor.row++;
	  }
	  cptr++;
	  break;
	default:
	  if(*cptr >= ' ' && *cptr != 127) {  // 127 = delete
	    if(showlen < window::ScrCols)
	      showlen++;
	    *str++ = *cptr++;
	  } else {
	    sprintf((char *)numbuf, "%3.3o", (int)*cptr);
	    *str++ = '\\';
	    *str++ = numbuf[0];
	    *str++ = numbuf[1];
	    *str++ = numbuf[2];
	    if((showlen += 4) > window::ScrCols)
	      showlen = window::ScrCols;
	    cptr++;
	  }
	}
      }
    }
    *str = '\0';
    
    if(showlen == window::ScrCols && autowarp) {
      if(cursor.row < Scroll.end)
	cursor.row++;
      cursor.col = 1;
      printf("%s\n\r", buf);
      Scrcursor = cursor;
    } else {
      cursor.col = showlen + 1;
      if(cursor.col > window::ScrCols)
	cursor.col = window::ScrCols;
      
      printf("%s", buf);
      Scrcursor.col = cursor.col;
    }
    str = buf;
  }
  fflush(stdout);
}

void window::print(const unsigned char *cptr, const unsigned char *add)
{
  unsigned char buf[BUFFER_SIZE + 100], *str, tmpbuf[100], *tmp;
  unsigned char numbuf[100], *num, numbuf2[100], *num2;
  int showlen = 0, addlen = 0, i, decmode, WinEnd;

  if(window::ScrScroll.start != Scroll.start ||
     window::ScrScroll.end   != Scroll.end) {
    scroll_region(&Scroll);
  }
  
  if(window::Scrcursor.col != cursor.col ||
     window::Scrcursor.row != cursor.row) {
    move(&cursor);
  }
  
  check_color();
   
  WinEnd = WinSize.end;
  if(get_time_on())
    WinEnd--;
  
  addlen = str_len(add);
  if(addlen >= window::ScrCols)
    return;
    
  while(*cptr && (*cptr == '\r' || *cptr == '\n'))
    cptr++;
  
  str = buf;
  while(*cptr) {
    showlen = window::Scrcursor.col - 1;
    printf("%s", add);
    showlen += addlen;
    while((showlen < window::ScrCols || !autowarp) && *cptr) {
      if(*cptr == ESC) {
	*str = '\0';
	cursor.col = showlen + 1;
	if(cursor.col > window::ScrCols)
	  cursor.col = window::ScrCols;

	printf("%s", buf);
	Scrcursor.col = cursor.col;
	str = buf;
	
	tmp = tmpbuf;
	*tmp++ = *cptr++;
	if(*cptr == '[') {
	  *tmp++ = *cptr++;
	  
	  decmode = FALSE;
	  if(*cptr == '?') {
	    decmode = TRUE;
	    *tmp++ = *cptr++;
	  }
	  
	  num = numbuf;
	  while(isdigit(*cptr)) {
	    *num++ = *cptr;
	    *tmp++ = *cptr++;
	  }
	  *num = '\0';
	  
	  num2 = numbuf2;
	  if(*cptr == ';') {
	    *tmp++ = *cptr++;
	    while(isdigit(*cptr)) {
	      *num2++ = *cptr;
	      *tmp++  = *cptr++;
	    }
	  }
	  *num2 = '\0';
	  while(*cptr == ';') {
	    *tmp++ = *cptr++;
	    while(isdigit(*cptr)) {
	      *num2  = *cptr;
	      *tmp++ = *cptr++;
	    }
	  }
	  *tmp++ = *cptr;
	  *tmp = '\0';
	  
	  if(decmode) {
	    switch(*cptr) {
	    case 'l':
	      if(*numbuf)
		i = atoi((const char *)numbuf);
	      else
		i = 0;
	      switch(i) {
	      case 0:                   // IGNORED
		break;
	      case 7:                   // AUTOWARP OFF
		autowarp = FALSE;
		break;
	      }
	      cptr++;
	      break;
	    case 'h':
	      if(*numbuf)
		i = atoi((const char *)numbuf);
	      else
		i = 0;
	      switch(i) {
	      case 0:                   // IGNORED
		break;
	      case 7:                   // AUTOWARP ON
		autowarp = TRUE;
		break;
	      }
	      cptr++;
	      break;
	    default:
	      cptr++;
	      break;
	    }  /* switch */
	  } else {
	    switch(*cptr) {
	    case 'm':
	      printf("%s", tmpbuf);
	      color_to_bits(tmpbuf, colormode);
	      window::Scrcolormode[0] = colormode[0];
	      window::Scrcolormode[1] = colormode[1];
	      window::Scrcolormode[2] = colormode[2];
	      window::Scrcolormode[3] = colormode[3];
	      cptr++;
	      break;
	    default:
	      cptr++;
	      break;
	    }  /* switch */
	  }
 	} else {
	  if(*cptr)
	    cptr++;
	}
      } else {
	switch(*cptr) {
	case '\b':                       // backspace
	case '\r':
	case '\f':
	  cptr++;
	  break;
	case '\t':                       // tab
	  *str = '\0';
	  cursor.col = showlen + 1;
	  if(cursor.col > window::ScrCols)
	    cursor.col = window::ScrCols;
	  
	  printf("%s", buf);
	  Scrcursor.col = cursor.col;
	  str = buf;
	  
	  showlen++;
	  if(showlen % TABSIZE)
	    showlen += TABSIZE - (showlen % TABSIZE);
	  
	  if(showlen + 1 > window::ScrCols)
	    showlen = window::ScrCols;
	  
	  move(cursor.row, showlen);
	  cptr++;
	  break;
	case '\a':                       // bell
	  bell();
	  cptr++;
	  break;
	case '\n':
	  *str = '\0';
	  if(cursor.row < Scroll.end)
	    cursor.row++;
	  cursor.col = 1;
	  printf("%s\n\r", buf);
	  Scrcursor = cursor;
	  showlen = 0;
	  
	  str = buf;
	  while(*cptr && (*cptr == '\n' || *cptr == '\r'))
	    cptr++;
	  
	  if(*cptr) {
	    printf("%s", add);
	    showlen += addlen;
	  }
	  break;
	default:
	  if(*cptr >= ' ' && *cptr != 127) {  // 127 = delete
	    if(showlen < window::ScrCols)
	      showlen++;
	    *str++ = *cptr++;
	  } else {
	    sprintf((char *)numbuf, "%3.3o", (int)*cptr);
	    *str++ = '\\';
	    *str++ = numbuf[0];
	    *str++ = numbuf[1];
	    *str++ = numbuf[2];
	    if((showlen += 4) > window::ScrCols)
	      showlen = window::ScrCols;
	    cptr++;
	  }
	}
      }
    }
    *str = '\0';
    
    if((showlen == window::ScrCols && autowarp) || *buf) {
      if(cursor.row < Scroll.end)
	cursor.row++;
      cursor.col = 1;
      printf("%s\n\r", buf);
      Scrcursor = cursor;
    }
    str = buf;
  }
  fflush(stdout);
}

int window::change_time(const int delay)
{
  unsigned char buf[BUFFER_SIZE];
  session *ses, *nextses;
  
  if(delay < 0) {
    if(timedelay < 0) {
      return(FALSE);
    } else {
      active_ses->print((const uchar *)"#Time data OFF.\n\r");
      timedelay = delay;
      if(WinSize.end == Scroll.end + 1)
	Scroll.end++;
      
      redraw();
      for(ses = session::get_sessions(); ses; ses = nextses) {
	nextses = ses->get_next();
	if(ses->get_output() == this)
	  ses->send_naws();
      }
    }
  } else {
    sprintf((char *)buf, "#Time data ON. (updating every %d seconds.)\n\r", delay);
    if(timedelay < 0) {
      if(Scroll.start + 1 < Scroll.end ||
	 Scroll.end < WinSize.end) {
	active_ses->print(buf);
	timedelay = delay;
	if(WinSize.end == Scroll.end)
	  Scroll.end--;
	redraw();
	for(ses = session::get_sessions(); ses; ses = nextses) {
	  nextses = ses->get_next();
	  if(ses->get_output() == this)
	    ses->send_naws();
	}
      } else {
	active_ses->print((const uchar *)"#Too small window to clock line.\n\r");
	return(FALSE);
      }
    } else {
      active_ses->print(buf);
      if((timedelay != 0 && delay == 0) ||
	 (timedelay == 0 && delay != 0)) {
	timedelay = delay;
	print_time(TRUE);
      } else
	timedelay = delay;
    }
  }
  return(TRUE);
}

void window::print_time(const int all, const int bell)
{
  const uchar *colstr;
  unsigned char buf[256], buf2[256];
  int hours, mins, tmp, linenum;
  int everything;
  time_t timep, secs;
  struct position orgpos;
  struct tm *timedata;
  struct listnode *ln;
  static int cal_mins, time_mins;
  static session *lastses = 0;
  window *activ;
  
  if(active_ses) {
    if(timedelay > 0) {
      if(lastses != active_ses || all) {
	everything = TRUE;
	lastses = active_ses;
      } else {
	everything = FALSE;
      }
      linenum = WinSize.end;
      
      timep = time(0);
      
      orgpos = cursor;
      timedata = localtime(&timep);
      
      secs = timep - active_ses->get_time();
      hours = secs / (60*60);
      secs -= hours * (60*60);
      mins = secs / 60;
      secs -= mins * 60;
      
      if((ln = searchnode_list(active_ses->colors, (unsigned char *)"TIMELINEON")) != NULL) {
	printf("%s", ln->right);
	color_to_bits(ln->right, window::Scrcolormode);
      } else {
	printf("%s", TIMELINEON);
	color_to_bits(TIMELINEON, window::Scrcolormode);
      }
      
      if(everything || time_mins != timedata->tm_min) {
	time_mins = timedata->tm_min;
	move(linenum, 1);
	printf("Clock: %2.2d.%2.2d.19%2.2d %2.2d:%2.2d:%2.2d",
	       timedata->tm_mday,
	       timedata->tm_mon + 1,
	       timedata->tm_year,
	       timedata->tm_hour,
	       timedata->tm_min,
	       timedata->tm_sec);
      } else {
	move(linenum, 25);
	printf("%2.2d",
	       timedata->tm_sec);
      }

      if(everything || cal_mins != mins || lastactive != window::active ||
	 bell) {
	cal_mins = mins;
	lastactive = window::active;
	if((tmp = (window::ScrCols - 14) - 26) > 0) {
	  sprintf((char *)buf, "%%%d.%ds", tmp, tmp);
	  sprintf((char *)buf2, "%s%s(%s:%s) ",
		  bell ? " <BELL> " : "",
		  this == window::active ? "ACTIVE " : "",
		  name,
		  active_ses->get_name());
	  printf((const char *)buf, buf2);
	}
	printf("Time: %2.2d:%2.2d:%2.2d", hours, mins, (int)secs);
      } else {
	move(linenum, window::ScrCols - 1);
	printf("%2.2d", (int)secs);
      }
      cursor = orgpos;
      activ = window::get_active_win();
      activ->move(activ->get_cursor());
      
      fflush(stdout);
      activ->check_color();
    } else if(!timedelay) {
      if(all || lastactive != window::active || bell) {
	lastactive = window::active;
	
	orgpos = cursor;
	move(WinSize.end, 1);
	if((ln = searchnode_list(active_ses->colors, (unsigned char *)"TIMELINEON")) != NULL) {
	  printf("%s", ln->right);
	  color_to_bits(ln->right, window::Scrcolormode);
	} else {
	  printf("%s", TIMELINEON);
	  color_to_bits(TIMELINEON, window::Scrcolormode);
	}
	sprintf((char *)buf, "%%-%d.%ds", window::ScrCols, window::ScrCols);
	sprintf((char *)buf2,"%s(%s:%s)%s",
		this == window::active ? "ACTIVE " : "",
		name,
		active_ses->get_name(),
		bell ? "  <BELL>" : "");
	printf((const char *)buf, buf2);
	
	cursor = orgpos;
	activ = window::get_active_win();
	activ->move(activ->get_cursor());
	
 	fflush(stdout);
	activ->check_color();
      }
    }
  } else {
    if(timedelay >= 0) {
      if(all || lastactive != window::active || bell) {
	lastactive = active;
	
	orgpos = cursor;
	move(WinSize.end, 1);
	colstr = TIMELINEON;
	
	sprintf((char *)buf, "%%s%%-%d.%ds", window::ScrCols, window::ScrCols);
	sprintf((char *)buf2,"%s(%s:(none))%s",
		this == window::active ? "ACTIVE " : "",
		name,
		bell ? "  <BELL>" : "");
	printf((char *)buf, colstr, buf2);
	color_to_bits(colstr, window::Scrcolormode);
 
	cursor = orgpos;
	activ = window::get_active_win();
	activ->move(activ->get_cursor());
	
	activ->check_color();
	
	fflush(stdout);
      }
    }
  }
}

void window::print_win_hist_line(const int all, unsigned char *text)
{
  unsigned char buffer[20];
  int tmp, lines, num, WinEnd, i;
  static int lastkey = 0;
  position orgpos;
  
  extern void (*keys[256])(int, window *);
  void func_exit_win_hist(int prompt_len, window *activ);
  
  WinEnd = WinSize.end;
  if(get_time_on())
    WinEnd--;
  
  check_color();
  
  if(window_hist_on) {
    orgpos = cursor;

    lines = WinEnd - WinSize.start;
    
    for(i = 0; i < 'z' - 'a' + 1; i++)
      if(keys[i + 1] == func_exit_win_hist)
	break;

    if(i < 'z' - 'a' + 1) {
      sprintf((char *)buffer, "ctrl-%c", i + 'a');
    } else {
      sprintf((char *)buffer, "no key");
    }

    num = window_hist_on->get_hist_num();
    tmp = (MAX_SES_HIST > num ? num : MAX_SES_HIST) -
      (lines > num ? num : lines);
    
    if((master_table.inputbuf_pos && window::active == this) || text) {
      if(all || lastkey != i) {
	lastkey = i;
	move(WinEnd, 1);               /* last line begin */
	
	if(text) {
	  printf("[%s to exit] (pos: %3d/%3d) ERROR: %s%s",
		 buffer,
		 window_hist_pos, tmp, text, CEL);
	} else {
	  printf("[%s to exit] (pos: %3d/%3d) S: %s%s",
		 buffer,
		 window_hist_pos, tmp, master_table.inputbuf, CEL);
	}
      } else {
	move(WinEnd, 24);              /* last line begin */
	if(text) {
	  printf("%3d/%3d) ERROR: %s%s",
		 window_hist_pos, tmp, text, CEL);
	} else {
	  printf("%3d/%3d) S: %s%s",
		 window_hist_pos, tmp, master_table.inputbuf, CEL);
	}
      }
    } else {
      if(all || lastkey != i) {
	lastkey = i;
	move(WinEnd, 1);               /* last line begin */
	printf("[%s to exit] (pos: %3d/%3d)%s",
	       buffer,
	       window_hist_pos, tmp, CEL);
      } else {
	move(WinEnd, 24);              /* last line begin */
	printf("%3d/%3d)%s", window_hist_pos, tmp, CEL);
      }
    }
    move(&orgpos);
    fflush(stdout);
  }
}

void window::print_prompt(const int all, const int log)
{
  int lines, pos, prompt_len = 0;
  session *ses = 0;
  window *win = 0;
  
  check_color();

  if(active_ses) {
    prompt_len = active_ses->get_prompt_len();
  } else {
    for(win = window::windows; win; win = win->next) {
      if(win->get_active_ses() &&
	 win->get_active_ses()->get_output() == this) {
	ses = win->get_active_ses();
	break;
      }
    }
    if(!ses) {
      for(ses = master_table.main_session; ses; ses = ses->get_next())
	if(ses->get_output() == this)
	  break;
    }
    if(ses)
      prompt_len = ses->get_prompt_len();
  }
  
  if(all && ptexrem) {
    if(active_ses) {
      if(active_ses->line_mode) {
	if(log) {
	  active_ses->print(active_ses->get_prompt());
	  if(this != active_ses->get_output())
	    print(active_ses->get_prompt());
	} else
	  print(active_ses->get_prompt());
      }
    } else if(!log) {
      if(ses && ses->line_mode)
	print(ses->get_prompt());
    }
    ptexrem = FALSE;
  }

  if(this == window::active && active_ses) {
    if(master_table.utexrem) {
      if(INPUTPOS) {
	if(!window_hist_on && active_ses->term_echoing) {
	  print(INPUTBUF);
	  
	  if(INPUTPOS > CURSOR) {
	    lines = ((prompt_len + INPUTPOS) / window::ScrCols) -
	      ((prompt_len + CURSOR) / window::ScrCols);
	    
	    pos = (prompt_len + CURSOR) % window::ScrCols;
	    
	    move(cursor.row - lines, pos + 1);
	  }
	}
      }
      master_table.utexrem = FALSE;
    }
  }

  fflush(stdout);
}

void window::remove_user_text()
{
  int lines;
  int prompt_len;
  prompt_len = active_ses ? active_ses->get_prompt_len() : 0;
  
  check_color();
  
  if(this == window::active) {
    if(!window_hist_on) {
      if(INPUTPOS && !master_table.utexrem) {
	if(CURSOR) {
	  lines = (prompt_len + CURSOR) / window::ScrCols;
	  
	  move(cursor.row - lines, prompt_len + 1);
	} else {
	  if(window::Scrcursor.col != cursor.col ||
	     window::Scrcursor.row != cursor.row)
	    move(&cursor);
	}
	clearend();
      }
    }
    master_table.utexrem = TRUE;
  }
}

void window::hist_movebackward()
{
  int lines, WinEnd, line, len, tmp, pos;

  check_color();
  
  WinEnd = WinSize.end;
  if(get_time_on())
    WinEnd--;
  
  lines = WinEnd - WinSize.start;
  
  if(lines > window_hist_on->get_hist_num())
    lines = window_hist_on->get_hist_num();
  
  tmp = WinSize.start + (lines - 1) - cursor.row;
  if(tmp < 0)
    tmp = 0;
  
  pos = window_hist_pos + tmp + 1;

  if(cursor.col > 1) {
    line = window_hist_on->get_hist_end() - pos;
    if(line < 0)
      line += MAX_SES_HIST;
    
    len = strlen((const char *)window_hist_on->get_scrhistory(line));
    
    cursor.col--;
    if(cursor.col > len)
      cursor.col = len;
    
    if(window::Scrcursor.col != cursor.col + 1 ||
       window::Scrcursor.row != cursor.row) {
      move(&cursor);
    } else {
      printf("%s", BACKWARD);
      window::Scrcursor.col = cursor.col;
    }
  } else {
    if(pos < window_hist_on->get_hist_num()) {
      pos++;
      line = window_hist_on->get_hist_end() - pos;
      if(line < 0)
	line += MAX_SES_HIST;
      
      cursor.col = strlen((const char *)window_hist_on->get_scrhistory(line));
      
      if(cursor.row > WinSize.start) {
	cursor.row--;
	move(&cursor);
      } else {
	hist_backward(1);
      }
    }
  }
}

void window::hist_moveforward()
{
  int lines, WinEnd, line, len, tmp, pos;

  WinEnd = WinSize.end;
  if(get_time_on())
    WinEnd--;
  
  lines = WinEnd - WinSize.start;
  
  if(lines > window_hist_on->get_hist_num())
    lines = window_hist_on->get_hist_num();
  
  tmp = WinSize.start + (lines - 1) - cursor.row;
  if(tmp < 0)
    tmp = 0;
  pos = window_hist_pos + tmp + 1;
  
  line = window_hist_on->get_hist_end() - pos;
  if(line < 0)
    line += MAX_SES_HIST;
  
  len = strlen((const char *)window_hist_on->get_scrhistory(line));
  if(len > window::get_cols())
    len = window::get_cols();

  check_color();

  if(cursor.col < len) {
    cursor.col++;
    
    if(window::Scrcursor.col != cursor.col - 1 ||
       window::Scrcursor.row != cursor.row) {
      move(&cursor);
    } else {
      printf("%s", FORWARD);
      window::Scrcursor.col = cursor.col;
    }
  } else {
    if(pos > 0) {
      cursor.col = 1;

      if(cursor.row < WinSize.start + (lines - 1)) {
	cursor.row++;
	move(&cursor);
      } else {
	hist_forward(1);
      }
    }
  }
}

void window::hist_backward(int amount)
{
  int lines, pos, WinEnd, num;
  position orgpos;

  if(window_hist_on) {
    WinEnd = WinSize.end;
    if(get_time_on())
      WinEnd--;

    lines = WinEnd - WinSize.start;
    
    if(cursor.row > WinSize.start) {
      num = cursor.row - WinSize.start;
      if(num > amount)
	num = amount;
      
      if(window::Scrcursor.row != cursor.row ||
	 window::Scrcursor.col != cursor.col) {
	cursor.row -= num;
	move(&cursor);
      } else {
	cursor.row -= num;
	printf("\033[%dA", num);
      }
      amount -= num;
    }
    
    if(window_hist_pos + lines < window_hist_on->get_hist_num() && amount){
      
      if(window_hist_pos + lines + amount > window_hist_on->get_hist_num())
	amount = window_hist_on->get_hist_num() - window_hist_pos - lines;
      
      orgpos = cursor;
      
      if(amount < lines) {
	if(window::ScrScroll.start != WinSize.start ||
	   window::ScrScroll.end   != WinEnd - 1)
	  scroll_region(WinSize.start, WinEnd - 1);
	
	check_color();
	
	for(num = 0; num < amount; num++) {
	  window_hist_pos++;
	  pos = (window_hist_on->get_hist_end() -
		 (window_hist_pos + lines));
	  if(pos < 0)
	    pos += MAX_SES_HIST;
	  
	  /* 1 line down all lines */
	  move(WinSize.start, 1);
	  printf("%s", INSERTLINE);
	  
	  printf("%s", window_hist_on->get_scrhistory(pos));
	}
	cursor = orgpos;
	print_win_hist_line(FALSE);
      } else {
	window_hist_pos += amount;
	redraw(-1);
      }
    }
  }
}

void window::hist_forward(int amount)
{
  int pos, WinEnd, num, lines;
  position orgpos;

  if(window_hist_on) {
    WinEnd = WinSize.end;
    if(get_time_on())
      WinEnd--;

    lines = WinEnd - 1 - WinSize.start;
    if(lines > window_hist_on->get_hist_num() - 1)
      lines = window_hist_on->get_hist_num() - 1;
    
    if(cursor.row < WinSize.start + lines) {
      num = (WinSize.start + lines) - cursor.row;
      if(num > amount)
	num = amount;
      
      if(window::Scrcursor.row != cursor.row ||
	 window::Scrcursor.col != cursor.col) {
        cursor.row += num;
        move(&cursor);
      } else {
        cursor.row += num;
        printf("\033[%dB", num);
      }
      amount -= num;
    }
    
    if(window_hist_pos > 0 && amount) {
      if(window_hist_pos < amount)
	amount = window_hist_pos;
      
      orgpos = cursor;
      
      if(amount < WinEnd - WinSize.start) {
	if(window::ScrScroll.start != WinSize.start ||
	   window::ScrScroll.end   != WinEnd - 1)
	  scroll_region(WinSize.start, WinEnd - 1);
	
	if(window::Scrcursor.col != window::ScrCols ||
	   window::Scrcursor.row != WinEnd - 1)
	  move(WinEnd - 1, window::ScrCols);
	
	check_color();
	
	for(num = 0; num < amount; num++) {
	  pos = window_hist_on->get_hist_end() - window_hist_pos;
	  if(pos < 0)
	    pos += MAX_SES_HIST;
	  window_hist_pos--;
	  printf("\n\r%s", window_hist_on->get_scrhistory(pos));
	}
	cursor = orgpos;
	print_win_hist_line(FALSE);
      } else {
	window_hist_pos -= amount;
	redraw(1);
      }
    }
  }
}

void window::hist_home()
{
  int WinEnd, pos;

  if(window_hist_on) {
    WinEnd = WinSize.end;
    if(get_time_on())
      WinEnd--;
    
    WinEnd--;
    pos = WinEnd - WinSize.start;
    if(pos > window_hist_on->get_hist_num() - 1)
      pos = window_hist_on->get_hist_num() - 1;
    
    if(window_hist_pos) {
      cursor.row = WinSize.start + pos;
      cursor.col = 1;
      
      window_hist_pos = 0;
      redraw();
    } else {
      move(WinSize.start + pos, 1);
    }
  }
}

void window::hist_end()
{
  int lines, pos = 0, WinEnd, line, len;
  
  if(window_hist_on) {
    WinEnd = WinSize.end;
    if(get_time_on())
      WinEnd--;

    lines = WinEnd - WinSize.start;
    pos = window_hist_on->get_hist_num() - lines;
    if(pos < 0)
      pos = 0;
    
    line = window_hist_on->get_hist_end() - window_hist_on->get_hist_num();
    if(line < 0)
      line += MAX_SES_HIST;
    
    len = strlen((const char *)window_hist_on->get_scrhistory(line));
    if(len > window::get_cols())
      len = window::get_cols();
    
    if(window_hist_pos != pos) {
      cursor.row = WinSize.start;
      cursor.col = len;
      window_hist_pos = pos;
      redraw();
    } else {
      move(WinSize.start, len);
    }
  }
}

void window::hist_pageup()
{
  int lines, WinEnd;

  if(window_hist_on) {
    WinEnd = WinSize.end;
    if(get_time_on())
      WinEnd--;
    
    lines = WinEnd - WinSize.start;
    
    hist_backward(lines);
  }
}

void window::hist_pagedown()
{
  int lines, WinEnd;
  
  if(window_hist_on) {
    WinEnd = WinSize.end;
    if(get_time_on())
      WinEnd--;
    
    lines = WinEnd - WinSize.start;
    
    hist_forward(lines);
  }
}

int window::hist_search(const unsigned char *str, const int forward)
{
  int lines, endpos, pos, found, num, i, line, WinEnd, tmp, col;

  if(window_hist_on) {
    WinEnd = WinSize.end;
    if(get_time_on())
      WinEnd--;
    
    lines = WinEnd - WinSize.start;
    
    if(lines > window_hist_on->get_hist_num())
      lines = window_hist_on->get_hist_num();
    
    tmp = WinSize.start + (lines - 1) - cursor.row;
    if(tmp < 0)
      tmp = 0;
    
    pos = window_hist_pos + tmp + 1;
    col = cursor.col - 1 + forward;
    
    endpos = window_hist_on->get_hist_num();
    
    found = FALSE;
    for(num = 0; pos <= endpos; pos++, num++) {
      line = window_hist_on->get_hist_end() - pos;
      if(line < 0)
	line += MAX_SES_HIST;
      
      if(col > 0) {
	if(col > (tmp = strlen((const char *)window_hist_on->get_scrhistory(line)))) {
	  col = 0;
	  continue;
	}
      }
      
      for(i = col; *(window_hist_on->get_scrhistory(line) + i) != '\0' ; i++)
	if(is_abrev(window_hist_on->get_scrhistory(line) + i, str)) {
	  found = TRUE;
	  col = i + 1;
	  break;
	}
      if(found)
	break;
      
      col = 0;
    }
    if(!found) {
      print_win_hist_line(FALSE, (unsigned char *)"Couldn't find such string");
      return(FALSE);
    }
    
    if(col > window::get_cols())
      col = window::get_cols();
    
    cursor.col = col;
    hist_backward(num);
  }
  return(TRUE);
}

void window::hist_set_mark()
{
  int lines, num, line, pos, len;
  position orgpos;
  struct listnode *ln;
  
  if(window_hist_on) {
    num = window_hist_on->get_hist_num();
    
    lines = WinSize.end - WinSize.start;
    if(get_time_on())
      lines--;
    
    if(num > lines)
      num = lines;

    orgpos = cursor;

    if(window_hist_on->get_hist_markp() > window_hist_pos &&
       window_hist_on->get_hist_markp() <= window_hist_pos + num) {
      
      move(WinSize.start + window_hist_pos + num -
	   window_hist_on->get_hist_markp(),
	   window_hist_on->get_hist_markc());
      
      check_color();
      
      line = window_hist_on->get_hist_end() - window_hist_on->get_hist_markp();
      if(line < 0)
	line += MAX_SES_HIST;
      
      printf("%c", *(window_hist_on->get_scrhistory(line) +
		     window_hist_on->get_hist_markc() - 1));
      
      move(&orgpos);
    }
    if((ln = searchnode_list(window_hist_on->colors, (unsigned char *)"HISTMARK")) != NULL) {
      printf("%s", ln->right);
      color_to_bits(ln->right, window::Scrcolormode);
    } else {
      printf("%s", TIMELINEON);
      color_to_bits(TIMELINEON, window::Scrcolormode);
    }
    pos = WinSize.start + window_hist_pos + num - cursor.row;
    
    line = window_hist_on->get_hist_end() - pos;
    if(line < 0)
      line += MAX_SES_HIST;

    len = strlen((const char *)window_hist_on->get_scrhistory(line));
    if(orgpos.col > len) {
      orgpos.col = len;
      move(&orgpos);
    }
    printf("%c", *(window_hist_on->get_scrhistory(line) + cursor.col - 1));
    
    window_hist_on->set_hist_mark(pos, cursor.col);
    check_color();
    move(&orgpos);
  }
}

void window::hist_copy_clip()
{
  unsigned char buffer[BUFFER_SIZE];
  int cursorpos, lines, num, line, len, amount, i, strpos;
  int startpos, startcol, endpos, endcol, tmprow;
  struct listnode *ln;
  position orgpos;
  
  if(window_hist_on) {
    if(master_table.clip) {
      delete[] master_table.clip;
      master_table.clip = 0;
    }
    num = window_hist_on->get_hist_num();
    
    lines = WinSize.end - WinSize.start;
    if(get_time_on())
      lines--;
    
    if(num > lines)
      num = lines;
    
    cursorpos = WinSize.start + window_hist_pos + num - cursor.row;
    
    line = window_hist_on->get_hist_end() - cursorpos;
    if(line < 0)
      line += MAX_SES_HIST;
    
    len = strlen((const char *)window_hist_on->get_scrhistory(line));
    if(cursor.col > len) {
      cursor.col = len;
      move(&cursor);
    }

    if(cursorpos == window_hist_on->get_hist_markp()) {
      if(window_hist_on->get_hist_markc() < cursor.col) {
	startcol = window_hist_on->get_hist_markc();
	endcol   = cursor.col;
      } else {
	startcol = cursor.col;
	endcol   = window_hist_on->get_hist_markc();
      }
      amount = endcol - startcol + 2;

      master_table.clip = new unsigned char[amount];
      
      if(!master_table.clip) {
	print_win_hist_line(FALSE, (unsigned char *)"OUT OF MEMORY");
	return;
      }      
      strpos = 0;
      for(i = startcol - 1; i < endcol; i++)
	*(master_table.clip + strpos++) = window_hist_on->get_scrhistory(line)[i];
      *(master_table.clip + strpos++) = '\0';

      orgpos = cursor;
      if(cursor.col != startcol)
	move(cursor.row, startcol);
      
      if((ln = searchnode_list(window_hist_on->colors, (unsigned char *)"HISTMARK")) != NULL) {
	printf("%s", ln->right);
	color_to_bits(ln->right, window::Scrcolormode);
      } else {
	printf("%s", TIMELINEON);
	color_to_bits(TIMELINEON, window::Scrcolormode);
      }
      printf("%s", master_table.clip);
      move(&orgpos);
      check_color();
    } else {
      if(cursorpos < window_hist_on->get_hist_markp()) {
	startpos = window_hist_on->get_hist_markp();
	startcol = window_hist_on->get_hist_markc();
	endpos = cursorpos;
	endcol = cursor.col;
      } else {
	startpos = cursorpos;
	startcol = cursor.col;
	endpos = window_hist_on->get_hist_markp();
	endcol = window_hist_on->get_hist_markc();
      }
      
      line = window_hist_on->get_hist_end() - startpos;
      if(line < 0)
	line += MAX_SES_HIST;
      amount = strlen((const char *)window_hist_on->get_scrhistory(line)) - startcol + 1;
      
      for(i = startpos - 1;
	  i > endpos;
	  i--) {
	line = window_hist_on->get_hist_end() - i;
	if(line < 0)
	  line += MAX_SES_HIST;
	
	amount += strlen((const char *)window_hist_on->get_scrhistory(line));
      }
      amount += endcol + 1;
      
      if(amount >= BUFFER_SIZE) {
	print_win_hist_line(FALSE, (unsigned char *)"Too Long String To Clip");
	return;
      }
      
      master_table.clip = new unsigned char[amount];
      if(!master_table.clip) {
	print_win_hist_line(FALSE, (unsigned char *)"OUT OF MEMORY");
	return;
      }
      
      strpos = 0;
      
      orgpos = cursor;
      
      if((ln = searchnode_list(window_hist_on->colors, (unsigned char *)"HISTMARK")) != NULL) {
        printf("%s", ln->right);
        color_to_bits(ln->right, window::Scrcolormode);
      } else {
        printf("%s", TIMELINEON);
        color_to_bits(TIMELINEON, window::Scrcolormode);
      }
      
      line = window_hist_on->get_hist_end() - startpos;
      if(line < 0)
        line += MAX_SES_HIST;
      
      strpos = str_cat(master_table.clip, strpos, amount,
		       window_hist_on->get_scrhistory(line) + startcol - 1);
      
      if(startpos > window_hist_pos && startpos <= window_hist_pos + num) {
	tmprow = WinSize.start + window_hist_pos + num - startpos;
	if(cursor.row != tmprow || cursor.col != startcol)
	  move(tmprow, startcol);
	printf("%s", window_hist_on->get_scrhistory(line) + startcol - 1);
      }
      
      for(i = startpos - 1;
	  i > endpos;
	  i--) {
	line = window_hist_on->get_hist_end() - i;
	if(line < 0)
	  line += MAX_SES_HIST;
	
	strpos = str_cat(master_table.clip, strpos, amount,
			 window_hist_on->get_scrhistory(line));
	
	if(i > window_hist_pos && i <= window_hist_pos + num) {
	  move(WinSize.start + window_hist_pos + num - i, 1);
	  printf("%s", window_hist_on->get_scrhistory(line));
	}
      }
      
      line = window_hist_on->get_hist_end() - endpos;
      if(line < 0)
	line += MAX_SES_HIST;
      
      len = strlen((const char *)window_hist_on->get_scrhistory(line));
      
      for(i = 0; i < endcol; i++) {
	*(master_table.clip + strpos++) = window_hist_on->get_scrhistory(line)[i];
	*(buffer + i) = window_hist_on->get_scrhistory(line)[i];
      }
      *(master_table.clip + strpos++) = '\0';
      *(buffer + i) = '\0';
      
      if(endpos > window_hist_pos && endpos <= window_hist_pos + num) {
	move(WinSize.start + window_hist_pos + num - endpos, 1);
	printf("%s", buffer);
      }
      
      move(&orgpos);
      check_color();
    }
    fflush(stdout);
    window_hist_clip = TRUE;
  }
}

void window::hist_rem_clip_color()
{
  unsigned char buffer[BUFFER_SIZE];
  int cursorpos, lines, num, line, len, i, strpos;
  int startpos, startcol, endpos, endcol, tmprow;
  position orgpos;
  struct listnode *ln;
  
  if(window_hist_on && window_hist_clip) {
    window_hist_clip = FALSE;

    num = window_hist_on->get_hist_num();
    
    lines = WinSize.end - WinSize.start;
    if(get_time_on())
      lines--;
    
    if(num > lines)
      num = lines;
    
    cursorpos = WinSize.start + window_hist_pos + num - cursor.row;
    
    line = window_hist_on->get_hist_end() - cursorpos;
    if(line < 0)
      line += MAX_SES_HIST;
    
    len = strlen((const char *)window_hist_on->get_scrhistory(line));
    if(cursor.col > len) {
      cursor.col = len;
      move(&cursor);
    }
    
    check_color();
    
    orgpos = cursor;
    
    if(cursorpos == window_hist_on->get_hist_markp()) {
      if(window_hist_on->get_hist_markc() < cursor.col) {
	startcol = window_hist_on->get_hist_markc();
	endcol   = cursor.col;
      } else {
	startcol = cursor.col;
	endcol   = window_hist_on->get_hist_markc();
      }
      strpos = 0;
      for(i = startcol - 1; i < endcol; i++)
	*(buffer + strpos++) = window_hist_on->get_scrhistory(line)[i];
      *(buffer + strpos) = '\0';
      
      if(cursor.col != startcol)
	move(cursor.row, startcol);
      
      printf("%s", buffer);
    } else {
      if(cursorpos < window_hist_on->get_hist_markp()) {
	startpos = window_hist_on->get_hist_markp();
	startcol = window_hist_on->get_hist_markc();
	endpos = cursorpos;
	endcol = cursor.col;
      } else {
	startpos = cursorpos;
	startcol = cursor.col;
	endpos = window_hist_on->get_hist_markp();
	endcol = window_hist_on->get_hist_markc();
      }
      strpos = 0;
      
      if(startpos > window_hist_pos && startpos <= window_hist_pos + num) {
	line = window_hist_on->get_hist_end() - startpos;
	if(line < 0)
	  line += MAX_SES_HIST;
	
	strpos = str_cat(buffer, strpos, sizeof(buffer),
			 window_hist_on->get_scrhistory(line) + startcol - 1);
	
	tmprow = WinSize.start + window_hist_pos + num - startpos;
	if(cursor.row != tmprow || cursor.col != startcol)
	  move(tmprow, startcol);
	printf("%s", window_hist_on->get_scrhistory(line) + startcol - 1);
      }
      
      for(i = startpos - 1;
	  i > endpos;
	  i--) {
	if(i > window_hist_pos && i <= window_hist_pos + num) {
	  line = window_hist_on->get_hist_end() - i;
	  if(line < 0)
	    line += MAX_SES_HIST;
	  
	  move(WinSize.start + window_hist_pos + num - i, 1);
	  printf("%s", window_hist_on->get_scrhistory(line));
	}
      }
      
      if(endpos > window_hist_pos && endpos <= window_hist_pos + num) {
	line = window_hist_on->get_hist_end() - endpos;
	if(line < 0)
	  line += MAX_SES_HIST;
	
	len = strlen((const char *)window_hist_on->get_scrhistory(line));
	
	for(i = 0; i < endcol; i++) {
	  *(buffer + i) = window_hist_on->get_scrhistory(line)[i];
	}
	*(buffer + i) = '\0';
	
	move(WinSize.start + window_hist_pos + num - endpos, 1);
	printf("%s", buffer);
      }
    }
    
    if(window_hist_on->get_hist_markp() > window_hist_pos &&
       window_hist_on->get_hist_markp() <= window_hist_pos + num) {
      
      move(WinSize.start + window_hist_pos + num -
	   window_hist_on->get_hist_markp(),
	   window_hist_on->get_hist_markc());
      
      if((ln = searchnode_list(window_hist_on->colors, (unsigned char *)"HISTMARK")) != NULL) {
        printf("%s", ln->right);
        color_to_bits(ln->right, window::Scrcolormode);
      } else {
        printf("%s", TIMELINEON);
        color_to_bits(TIMELINEON, window::Scrcolormode);
      }
      
      line = window_hist_on->get_hist_end() - window_hist_on->get_hist_markp();
      if(line < 0)
	line += MAX_SES_HIST;
      
      printf("%c", *(window_hist_on->get_scrhistory(line) +
		     window_hist_on->get_hist_markc() - 1));
      
      check_color();
    }
    move(&orgpos);
    
    fflush(stdout);
  }
}

int window::start_win_history(session *ses)
{
  int WinEnd;
  
  if(!window_hist_on) {
    WinEnd = WinSize.end;
    if(get_time_on())
      WinEnd--;

    if(WinEnd > WinSize.start) {
      remove_user_text();
      *INPUTBUF = '\0';
      INPUTPOS = 0;
      CURSOR = 0;
      print_prompt(FALSE);
      
      window_hist_on = ses;
      redraw();
      return(TRUE);
    } else {
      ses->print((const uchar *)"#Too small window for history.\n\r");
    }
  }
  return(FALSE);
}

void window::exit_win_history()
{
  if(window_hist_on) {
    window_hist_pos = 0;
    window_hist_on  = 0;

    redraw();
  }
}

void window::max_size()
{
  int tmp;
  session *ses, *nextses;
  
  tmp = WinSize.end - Scroll.end;
  if(next) {
    WinSize.end = next->WinSize.start - 1;
  } else {
    WinSize.end = window::get_rows();
  }
  Scroll.end  = WinSize.end - tmp;
  redraw();
  for(ses = session::get_sessions(); ses; ses = nextses) {
    nextses = ses->get_next();
    if(ses->get_output() == this)
      ses->send_naws();
  }
}

int window::change_size(const int newsize)
{
  int oldsize = WinSize.end - WinSize.start + 1;
  int sizediff, newlines = 0, lines;
  window *prewin = 0;
  session *ses, *nextses;

  assert(newsize >= 3);
  
  if(newsize != oldsize) {
    clear();

    sizediff = WinSize.end - Scroll.end;
    
    if(newsize > oldsize) {               // bigger window
      newlines = newsize - oldsize;
      
      if(next) {
	lines = next->get_size()->end - next->get_size()->start + 1;
	if(lines - 3 > 0) {
	  if(newlines >= lines - 3) {
	    WinSize.end = next->get_size()->end - 3;
	    newlines -= lines - 3;
	  } else {
	    WinSize.end += newlines;
	    newlines = 0;
	  }
	  Scroll.end  = WinSize.end - sizediff;
	  next->WinSize.start = WinSize.end + 1;
	  next->Scroll.start  = next->WinSize.start;
	  next->redraw();
	  
	  for(ses = session::get_sessions(); ses; ses = nextses) {
	    nextses = ses->get_next();
	    if(ses->get_output() == next)
	      ses->send_naws();
	  }
	}
      } else {
	lines = window::ScrRows - WinSize.end;
	if(lines > 0) {
	  if(newlines >= lines) {
	    WinSize.end = window::ScrRows;
	    newlines -= lines;
	  } else {
	    WinSize.end += newlines;
	    newlines = 0;
	  }
	  Scroll.end  = WinSize.end - sizediff;
	}
      }
      if(newlines > 0) {
	if(window::windows != this) {
	  for(prewin = window::windows;
	      prewin->next && prewin->next != this;
	      prewin = prewin->get_next())
	    ;
	}
	if(prewin) {
	  lines = prewin->get_size()->end - prewin->get_size()->start + 1;

	  if(lines - 3 > 0) {
	    sizediff = prewin->WinSize.end - prewin->Scroll.end;
	    
	    if(newlines >= lines - 3) {
	      WinSize.start = prewin->WinSize.start + 3;
	      newlines -= lines - 3;
	    } else {
	      WinSize.start -= newlines;
	      newlines = 0;
	    }
	    Scroll.start  = WinSize.start;
	    prewin->WinSize.end = WinSize.start - 1;
	    prewin->Scroll.end  = prewin->WinSize.end - sizediff;
	    prewin->redraw();
	    for(ses = session::get_sessions(); ses; ses = nextses) {
	      nextses = ses->get_next();
	      if(ses->get_output() == prewin)
		ses->send_naws();
	    }
 	  }
	}
      }
    } else {                             // smaller window
      WinSize.end = WinSize.start + newsize - 1;
      Scroll.end  = WinSize.end - sizediff;
      
      if(next) {
	next->WinSize.start = WinSize.end + 1;
	next->Scroll.start  = next->WinSize.start;
	
	next->redraw();
	for(ses = session::get_sessions(); ses; ses = nextses) {
	  nextses = ses->get_next();
	  if(ses->get_output() == next)
	    ses->send_naws();
	}
      } else {
#ifdef TERMCAP
	if(MOVE) {
	  printf("%s%s", tgoto((const char *)MOVE, 0, WinSize.end), CES);
	} else {
#endif
	  printf("\033[%d;1H%s",
		 WinSize.end + 1, CES); // clear end of clear...
#ifdef TERMCAP
	}
#endif
	fflush(stdout);
      }
    }
    redraw();
    for(ses = session::get_sessions(); ses; ses = nextses) {
      nextses = ses->get_next();
      if(ses->get_output() == this)
	ses->send_naws();
    }
  }
  return(newlines);
}

void window::next_window_active(window *newact)
{
  window *win = 0;
  session *newactive, *ses;

  if(!newact) {
    if(next)
      win = next;
    else
      win = window::windows;
  } else
    win = newact;
  
  if(win != this) {
    remove_user_text();
    window::active = win;
    if(!win->active_ses) {
      newactive = master_table.main_session;
      if(newactive->get_input()->get_active_ses() == newactive) {
	for(ses = session::get_sessions(); ses; ses = ses->get_next())
	  if(ses->get_input() == newactive->get_input())
	    break;
	
	newactive->get_input()->set_active_ses(ses);
      }
      win->set_active_ses(newactive);
      newactive->set_input(win);
    }
    print_time(TRUE);
    win->print_time(TRUE);
    if(window_hist_on)
      print_win_hist_line(FALSE);
    if(win->window_hist_on) {
      win->print_win_hist_line(TRUE);
    } else {
      if(window::get_scrcursor()->col != win->get_cursor()->col ||
	 window::get_scrcursor()->row != win->get_cursor()->row)
	win->move(win->get_cursor());
    }
  }
}

void window::get_screen_size()
{
  unsigned char buf[1024], *bufpos;
  int col, row;
  winsize WSize;

  ioctl(0, TIOCGWINSZ, &WSize);
  
  ScrRows = WSize.ws_row;
  ScrCols = WSize.ws_col;
  
  if(ScrCols < 40 || ScrRows < 2) {
    nocbreak_echo();
    printf("%s", AUTOWARPON);
    printf("Screen Size: %dx%d\n\r", WSize.ws_col , WSize.ws_row);
    printf("Couldn't get screen size or too small screen!!!\n\r"
	   "Smallest allowed screen size is: 46x2\n\r");
    while(ScrCols < 40 || ScrRows < 2) {
      printf("Type screen size (default: %dx%d): ", 
	     default_chars_per_line, default_lines_in_scr);
      for(bufpos = buf; (*bufpos = getchar()) != '\n'; bufpos++)
	;
      *bufpos = '\0';
      if(sscanf((const char *)buf, " %d x %d", &col, &row) != 2) {
	printf("\n\rUsing defaults values.\n\r");
	ScrCols = default_chars_per_line;
	ScrRows = default_lines_in_scr;
      } else {
	if(col >= 40 && row >= 2) {
	  ScrCols = col;
	  ScrRows = row;
	  default_chars_per_line = col;
	  default_lines_in_scr = row;
	}
      }
    }
    cbreak_noecho();
  }
  printf("%s", AUTOWARPOFF);
}

/* ***************************************************
 *     function to get screen size                   *
 *************************************************** */

void get_screen_size(void)
{
  int oldrows;
  window *win, *nextwin;
  
  oldrows = window::get_rows();
  printf("%s", MAXSCROLLSIZE);
  window::get_screen_size();
  if(window::get_rows() != oldrows) {
    for(win = window::get_windows(); win; win = nextwin) {
      nextwin = win->get_next();
      if(win->get_size()->end > window::get_rows()) {
	if(window::get_rows() - win->get_size()->start < 2) {
	  delete win;
	}
      }
    }
    for(win = window::get_windows(); win->get_next(); win = win->get_next())
      ;
    win->max_size();
  }
  redraw_screen();
}

/* ********************************************************
 *   clear screen and print text again from scr history   *
 ******************************************************** */

void redraw_screen(void)
{
  window *tmp;
  
  printf("%s", CLEARSCREEN);
  fflush(stdout);
  
  for(tmp = window::get_windows(); tmp; tmp = tmp->get_next())
    tmp->redraw();
  
  tmp = window::get_active_win();
  tmp->scroll_region(tmp->get_scroll());
}

// **********************
//   print all prompts
// **********************

void prompt(void)
{
  window *win;
  
  for(win = window::get_windows(); win; win = win->get_next())
    win->print_prompt(TRUE);
}

/* **************************************
 *  cbreak and echo                     *
 ************************************** */

#ifndef _POSIX_VDISABLE
# ifdef SUN
#  include <sys/param.h>        /* pick up VDISABLE definition, mayby */
# endif
# ifdef VDISABLE
#  define _POSIX_VDISABLE VDISABLE
# else
#  define _POSIX_VDISABLE ((char)'\377')
# endif
#endif

void cbreak_noecho(void)
{
#ifdef TERMIOS
  struct termios cbreak_state;
  
  cbreak_state = original_state;
  cbreak_state.c_iflag &= ~(IXON | IXOFF);
  cbreak_state.c_lflag &= ~(ICANON | ISIG | ECHO);
  cbreak_state.c_cc[VMIN] = 1;
  cbreak_state.c_cc[VTIME] = 0;
  tcsetattr(0, TCSANOW, &cbreak_state);
#else
  struct sgttyb cbreak_state;
  struct tchars cbreak_chars;
  
  cbreak_state = original_state;
  cbreak_state.sg_flags |= CBREAK;
  cbreak_state.sg_flags |= RAW;
  cbreak_state.sg_flags &= ~ECHO;
  ioctl(0, TIOCSETP, &cbreak_state);
  cbreak_chars = original_chars;
  cbreak_chars.t_startc = _POSIX_VDISABLE;
  cbreak_chars.t_stopc  = _POSIX_VDISABLE;
  ioctl(0, TIOCSETC, &cbreak_chars);
#endif
}

void nocbreak_echo(void)
{
#ifdef TERMIOS
  tcsetattr(0, TCSANOW, &original_state);
#else
  ioctl(0, TIOCSETP, &original_state);
  ioctl(0, TIOCSETC, &original_chars);
#endif
}

// *********
//  termcap
// *********

const uchar *check_tertmcap_str(const char *str)
{
  if(!str)
    return(0);
  
  while(*str >= ' ')  // remove '.', numbers and '*' before control string
    str++;
  
  if(!*str)
    return(0);
  
  return((const uchar *)str);
}

void init_termcap()
{
#ifdef TERMCAP
  static char tbuf[1024];
  static char cbuf[1024];
  char *_tptr, *term;

  int len;

  if((term = getenv("TERM"))) {
    len = tgetent(cbuf, term);

    if(len > 0) {
      _tptr = tbuf;
 
      CEL         = check_tertmcap_str(tgetstr("ce", &_tptr));
      CES         = check_tertmcap_str(tgetstr("cd", &_tptr));
      CLEARSCREEN = check_tertmcap_str(tgetstr("cl", &_tptr));
      HOME        = check_tertmcap_str(tgetstr("ho", &_tptr));
      CUP         = check_tertmcap_str(tgetstr("up", &_tptr));
      CDOWN       = check_tertmcap_str(tgetstr("do", &_tptr));
      FORWARD     = check_tertmcap_str(tgetstr("nd", &_tptr));
      BACKWARD    = check_tertmcap_str(tgetstr("le", &_tptr));
      RESETCOLOR  = check_tertmcap_str(tgetstr("me", &_tptr));
      INSERTLINE  = check_tertmcap_str(tgetstr("al", &_tptr));
      VBELL       = check_tertmcap_str(tgetstr("vb", &_tptr));
      MOVE        = check_tertmcap_str(tgetstr("cm", &_tptr));
      
      TABSIZE     = tgetnum("it");

      add_key_string(check_tertmcap_str(tgetstr("kh",&_tptr)),func_home);
      add_key_string(check_tertmcap_str(tgetstr("kI",&_tptr)),
		     func_insert_clip);
      add_key_string(check_tertmcap_str(tgetstr("kH",&_tptr)),func_end);
      add_key_string(check_tertmcap_str(tgetstr("kP",&_tptr)),func_pageup);
      add_key_string(check_tertmcap_str(tgetstr("kN",&_tptr)),func_pagedown);
      add_key_string(check_tertmcap_str(tgetstr("ku",&_tptr)),func_arrowup);
      add_key_string(check_tertmcap_str(tgetstr("kd",&_tptr)),func_arrowdown);
      add_key_string(check_tertmcap_str(tgetstr("kr",&_tptr)),func_arrowright);
      add_key_string(check_tertmcap_str(tgetstr("kl",&_tptr)),func_arrowleft);
      add_key_string(check_tertmcap_str(tgetstr("k1",&_tptr)),func_mark);
      add_key_string(check_tertmcap_str(tgetstr("k2",&_tptr)),func_copy_clip);


    } else {
      if(len == 0) {
	fprintf(stderr, "No termcap entry for %s\n", term);
      } else {
	fprintf(stderr, "No /etc/termcap present!\n");
      }
    }
  } else {
    fprintf(stderr, "Environment variable TERM not set\n");
  }
#endif

  /* use vt100/ansi codes if didn't got from termcap */

  if(!CEL) CEL = (const uchar *)"\033[0K";    /* clear end of line         */
  if(!CSL) CSL = (const uchar *)"\033[1K";    /* clear start of line       */
  if(!CL)  CL  = (const uchar *)"\033[2K";    /* clear line                */
  if(!CES) CES = (const uchar *)"\033[0J";    /* clear end of screen       */
  if(!CSS) CSS = (const uchar *)"\033[1J";    /* clear start of screen     */
  if(!CLEARSCREEN) CLEARSCREEN = (const uchar *)"\033[H\033[2J"; /* clear screen              */
  if(!AUTOWARPON)  AUTOWARPON  = (const uchar *)"\033[?7h"; /* turns autowarp on         */
  if(!AUTOWARPOFF) AUTOWARPOFF = (const uchar *)"\033[?7l"; /* turns autowarp off        */
  if(!HOME)      HOME      = (const uchar *)"\033[H";
  if(!CUP)       CUP       = (const uchar *)"\033[A";
  if(!CDOWN)     CDOWN     = (const uchar *)"\033[B";
  if(!FORWARD)   FORWARD   = (const uchar *)"\033[C";
  if(!BACKWARD)  BACKWARD  = (const uchar *)"\033[D";
  if(!MAXSCROLLSIZE) MAXSCROLLSIZE = (const uchar *)"\033[r";
  if(!RESETCOLOR) RESETCOLOR = (const uchar *)"\033[0m";
  if(!INSERTLINE) INSERTLINE = (const uchar *)"\033[L";
  
  if(TABSIZE <= 0)  TABSIZE  = 8;
}

// ************************
//   command split window
// ************************

session *comm_split(const unsigned char *arg, session *ses)
{
  unsigned char buf[BUFFER_SIZE], name[BUFFER_SIZE];
  int value = 0, lines;
  window *win, *win2;
  session *sesptr;

  arg = get_arg_stop_spaces(arg, name);
  arg = get_arg_with_spaces(arg, buf);
  if(*buf) {
    value = string_to_number(buf, ses);
    
    win = ses->get_input();
    lines = win->get_size()->end - win->get_size()->start + 1;
    
    if(value < 3) {
      ses->print((const uchar *)"#You cannot make that small windows.\n\r");
    } else if(value > lines - 3) {
      ses->print((const uchar *)"#Too BIG new window\n\r");
    } else {
      win2 = find_window(name);
      if(win2) {
	ses->print((const uchar *)"#Already that named window\n\r");
	return(ses);
      }
      for(sesptr = master_table.main_session;
	  sesptr;
	  sesptr = sesptr->get_next())
	if(strcmp((const char *)sesptr->get_name(), (const char *)name) == 0) {
	  ses->print((const uchar *)"#Already that named session.\n\r");
	  return(ses);
	}
      
      ses->print((const uchar *)"#Ok.\n\r");
      ses->print_prompt(TRUE);
      win2 = new window(name, value, win);
      ses = window::get_active_win()->get_active_ses();
    }
  } else {
    ses->print((const uchar *)"#Usage: #split <NAME> <NEW WINDOW SIZE>\n\r");
  }
  return(ses);
}

// ***************************
//   command redirect output
// ***************************


session *comm_output(const unsigned char *arg, session *ses)
{
  unsigned char name[BUFFER_SIZE];
  window *win;
  
  arg = get_arg_with_spaces(arg, name);
  if(*name) {
    win = find_window(name);
    if(win && win != NULL_WIN) {
      ses->set_output(win);
      ses->print((const uchar *)"#Ok.\n\r");
      ses = ses->send_naws();
    } else {
      ses->print((const uchar *)"#NO such named window\n\r");
    }
  } else {
    ses->print((const uchar *)"#Usage: #output <WINDOW NAME>\n\r");
  }
  return(ses);
}

// **************************
//   command redirect input
// **************************

session *comm_input(const unsigned char *arg, session *ses)
{
  unsigned char name[BUFFER_SIZE];
  session *newact;
  window *win;
  
  arg = get_arg_with_spaces(arg, name);
  if(*name) {
    win = find_window(name);
    if(win && win != NULL_WIN) {
      if(win != ses->get_input()) {
	if(ses == ses->get_input()->get_active_ses()) {
	  for(newact = session::get_sessions();
	      newact;
	      newact = newact->get_next())
	    if(newact != ses && newact->get_input() == ses->get_input())
	      break;
	  ses->get_input()->set_active_ses(newact);
	}
	ses->set_input(win);
	win->set_active_ses(ses);
	window::set_active_win(win);
	if(!win->get_hist_on()) {
	  win->print((const uchar *)"\r");
	  win->set_ptexrem();
	}
      }
      ses->print((const uchar *)"#Ok.\n\r");
    } else {
      ses->print((const uchar *)"#NO such named window\n\r");
    }
  } else {
    ses->print((const uchar *)"#Usage: #input <WINDOW NAME>\n\r");
  }
  return(ses);
}

// *************************
//   command close window
// *************************

session *comm_close(const unsigned char *arg, session *ses)
{
  delete ses->get_input();
  redraw_screen();
  
  return(ses);
}

// ******************************
//   command window information
// ******************************

session *comm_wininfo(const unsigned char *arg, session *ses)
{
  unsigned char buf[BUFFER_SIZE], color[BUFFER_SIZE];
  unsigned char buf2[BUFFER_SIZE], tmp[4] = { 0x01, 0, 0, 0};
  window *win;

  bits_to_color(tmp, window::get_scrcolormode(), color);
  if(*color)
    color_to_text(color, buf2, FALSE);
  else
    sprintf((char *)buf2, "reset");

  for(win = window::get_windows(); win; win = win->get_next()) {
    bits_to_color(tmp, win->get_colormode(), color);
    if(*color)
      color_to_text(color, buf2, FALSE);
    else
      sprintf((char *)buf2, "reset");
    sprintf((char *)buf,
	    "#Window Name     : %s\n\r"
	    "#  Size          : %2d%s lines (Screen lines: %2d-%2d)\n\r"
	    "#  Scroll        : %2d      lines (Screen lines: %2d-%2d)\n\r"
	    "#  Cursor        : %dCol, %dRow\n\r"
	    "#  Colors:       : %s\n\r"
	    "#  Active session: %s\n\r",
	    win->get_name(),
	    (win->get_size()->end -
	     win->get_size()->start + 1 -
	     win->get_time_on()),
	    (win->get_time_on() ? " (+1)" : "     "),
	    win->get_size()->start,
	    win->get_size()->end,
	    (win->get_scroll()->end - win->get_scroll()->start + 1),
	    win->get_scroll()->start,
	    win->get_scroll()->end,
	    win->get_cursor()->col,
	    win->get_cursor()->row,
	    buf2,
	    (win->get_active_ses() ?
	     (char *)win->get_active_ses()->get_name() :
	     "(none)"));
    ses->print(buf);
  }
  sprintf((char *)buf, "#Session Input/Output windows: %s/%s\n\r",
	  ses->get_input()->get_name(),
	  ses->get_output()->get_name());
  ses->print(buf);
  return(ses);
}

// ********************
// command winsize code
// ********************

session *comm_winsize(const unsigned char *arg, session *ses)
{
  unsigned char buffer[BUFFER_SIZE];
  int maxsize, newsize = 0, lines;
  window *pre, *win;

  arg = get_arg_with_spaces(arg, buffer);
  if(*buffer) {
    win = ses->get_output();
    maxsize = win->get_size()->end - win->get_size()->start + 1;
    if(win->get_next()) {
      lines = (win->get_next()->get_size()->end -
	       win->get_next()->get_size()->start + 1);
      maxsize += lines - 3;
    } else {
      maxsize += window::get_rows() - win->get_size()->end;
    }
    if(win != window::get_windows()) {
      for(pre = window::get_windows();
	  pre->get_next() && pre->get_next() != win;
	  pre = pre->get_next())
	;
      if(pre) {
	lines = (pre->get_size()->end - pre->get_size()->start + 1);
	maxsize += lines - 3;
      }
    }
    if(stricmp((const uchar *)"min", buffer) == 0)
      newsize = 3;
    else if(stricmp((const uchar *)"max", buffer) == 0)
      newsize = maxsize;
    else
      newsize = string_to_number(buffer, ses);
    
    if(newsize < 3)
      ses->print((const uchar *)"#Too Small window size.\n\r");
    else if(newsize > maxsize)
      ses->print((const uchar *)"#Too Big window size.\n\r");
    else {
      ses->print((const uchar *)"#Ok.\n\r");
      ses->print_prompt(TRUE);
      if(win->change_size(newsize))
	ses->print((const uchar *)"#Hmmm... problems somewhere...:( (make bug report..)\n\r");
    }
  } else {
    ses->print((const uchar *)"#Usage: #winsize <SIZE | max | min>\n\r");
  }
 
  return(ses);
}
  
// *******************************
//  code to find window with name
// *******************************

window *find_window(const unsigned char *name)
{
  window *win;
  
  if(strcmp("null", (const char *)name) == 0)
    return(NULL_WIN);
  
  for(win = window::get_windows(); win; win = win->get_next()) {
    if(strcmp((const char *)win->get_name(), (const char *)name) == 0)
      return(win);
  }
  
  return(0);
}
